#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import pytest
import sys
from pathlib import Path

# from pypipegraph.testing.fixtures import new_pipegraph, pytest_runtest_makereport  # noqa:F401

root = Path(__file__).parent.parent
sys.path.append(str(root / "src"))



def test_version_is_correct():
    import configparser
    from pathlib import Path
    import mbf_fileformats

    c = configparser.ConfigParser()
    c.read(Path(__file__).parent.parent / "setup.cfg")
    version = c["metadata"]["version"]
    assert version == mbf_fileformats.__version__



import numpy as np
import os
import pandas as pd
import io

from .util import open_file


def wiggle_to_intervals(
    filenameOrFileObject, comment_char=None, chromosome_mangler=None
):
    """read a (variable/fixed step) wiggle file and turn it into a df of {chr, start,stop, score}.
    """
    fo = open_file(filenameOrFileObject)
    d = fo.read()
    modes = set()
    if b"variableStep" in d:
        modes.add("variableStep")
    if b"fixedStep" in d:
        modes.add("fixedStep")
    if b"Bed mode" in d:
        modes.add("bed")
    if len(modes) > 1:
        fo.seek(0, os.SEEK_SET)
        return wiggle_to_intervals_slow(
            filenameOrFileObject, comment_char, chromosome_mangler
        )
    lines = np.array(d.strip().split(b"\n"))
    del d
    if comment_char:
        lines = lines[
            ~np.char.startswith(lines, "track")
            & ~np.char.startswith(lines, comment_char)
        ]
    else:
        lines = lines[~np.char.startswith(lines, b"track")]
    if len(modes) == 0:
        if len(lines[0].split()) == 4:
            modes.add("bed")
    if len(modes) == 0:
        raise ValueError(
            "Did not know how to handle %s, no mode found" % filenameOrFileObject
        )
    mode = list(modes)[0]
    if chromosome_mangler is None:
        chromosome_mangler = lambda x: x  # NOQA
    if mode == "bed":
        s = io.BytesIO(b"\n".join(lines))
        recarray = np.loadtxt(
            s,
            dtype=[
                ("chr", "|S50"),
                ("start", np.int32),
                ("stop", np.uint32),
                ("score", np.float),
            ],
        )
        res = pd.DataFrame(
            {
                "chr": [chromosome_mangler(x) for x in recarray["chr"]],
                "start": recarray["start"],
                "stop": recarray["stop"],
                "score": recarray["score"],
            }
        )
        return res
    elif mode == "variableStep":
        borders = np.char.startswith(lines, "variableStep")
        # n umpy.array([x.startswith('variableStep') for x in lines],
        # dtype=np.bool)
        border_offsets = np.where(borders)[0]
        dfs_by_chr = []
        for ii in range(0, len(border_offsets) - 1):
            start = border_offsets[ii]
            stop = border_offsets[ii + 1]
            data_lines = lines[start + 1 : stop]
            row = lines[start]
            chr = row[row.find(b"chrom=") + len(b"chrom=") :]
            chr = chr[: chr.find(" ")]
            chr = chromosome_mangler(chr)
            if row.find(b"span") != -1:
                span = int(row[row.find(b"span=") + len(b"span=") :])
            else:
                span = 1
            just_one_row = len(data_lines) == 1
            data_stream = io.StringIO("\n".join(data_lines) + "\n")
            data_lines = np.loadtxt(data_stream)
            if just_one_row:
                data_lines = np.array([data_lines])
            try:
                starts = np.array(data_lines[:, 0], dtype=np.int)
            except IndexError:
                print(row)
                print(data_lines)
                print(data_lines.value)
                raise
            stops = starts + span
            scores = data_lines[:, 1]
            chrs = np.array([chr] * len(scores), dtype=np.object)
            dfs_by_chr.append(
                pd.DataFrame(
                    {"chr": chrs, "start": starts, "stop": stops, "score": scores}
                )
            )
        # .sort_by(('chr','start'), [True, True]) # we rely on the wiggle being sorte.d..
        return pd.concat(dfs_by_chr)
    elif mode == "fixedStep":
        borders = np.char.startswith(lines, "fixedStep")
        # borders = np.array([x.startswith('fixedStep') for x in lines], dtype=np.bool)
        border_offsets = np.where(borders)[0]
        dfs_by_chr = []
        for ii in range(0, len(border_offsets) - 1):
            data_lines = lines[start + 1 : stop]
            # parse header line...
            row = lines[start]
            chr = row[row.find(b"chrom=") + len(b"chrom=") :]
            chr = chr[: chr.find(" ")]
            chr = chromosome_mangler(chr)
            if row.find(b"span") != -1:
                span = int(row[row.find(b"span=") + len(b"span=") :])
            else:
                raise ValueError("No span in fixed step?!")
            if row.find(b"start") != -1:
                start = row[row.find(b"start=") + len(b"start=") :]
                start = int(start[: start.find(" ")])
            else:
                raise ValueError("No start in fixed step?!")
            if row.find(b"step") != -1:
                step = row[row.find(b"step=") + len(b"step=") :]
                step = int(step[: step.find(" ")])
            else:
                raise ValueError("No step in fixed step?!")
            if step != span:
                raise ValueError(
                    "Parser currently only supports step == span, and they were not equal"
                )
            data_lines = io.StringIO("\n".join(data_lines))
            data_lines = np.loadtxt(data_lines)
            scores = data_lines
            starts = np.array([start + step * ii for ii in range(0, len(scores))])
            stops = stops + span
            chrs = [chr] * len(scores)
            dfs_by_chr.append(
                pd.DataFrame(
                    {"chr": chrs, "start": starts, "stop": stops, "score": scores}
                )
            )
        return pd.concat(dfs_by_chr)
    else:
        raise NotImplementedError("Unknown Wiggle mode")


def wiggle_to_intervals_slow(
    filenameOrFileObject, comment_char=None, chromosome_mangler=None
):
    fo = open_file(filenameOrFileObject)
    result = {}
    mode = None
    chr = None
    span = None
    start = None
    stop = None
    lastScore = None
    data = {"start": [], "stop": [], "score": [], "chr": []}
    if chromosome_mangler is None:
        chromosome_mangler = lambda x: x  # NOQA
    bytes_so_far = 0
    for row in fo:
        bytes_so_far += len(row)
        if (comment_char and row.startswith(comment_char)) or row.startswith("track"):
            continue
        elif row.startswith("variableStep"):
            mode = "variable"
            chr = row[row.find("chrom=") + len("chrom=") :]
            chr = chr[: chr.find(" ")]
            if chr not in result:
                result[chr] = []
            if row.find("span") != -1:
                span = row[row.find("span=") + len("span=") :]
                span = int(span)
            else:
                span = 1
            start = None
            stop = None
            lastScore = None
            continue
        elif row.startswith("fixedStep"):
            mode = "fixed"
            chr = row[row.find("chrom=") + len("chrom=") :]
            chr = chr[: chr.find(" ")]
            if chr not in result:
                result[chr] = []
            if row.find("span") != -1:
                span = row[row.find("span=") + len("span=") :]
                span = int(span)
            else:
                raise ValueError("No span in fixed step?!")
            if row.find("start") != -1:
                start = row[row.find("start=") + len("start=") :]
                start = start[: start.find(" ")]
                start = int(start)
            else:
                raise ValueError("No start in fixed step?!")
            if row.find("step") != -1:
                step = row[row.find("step=") + len("step=") :]
                step = step[: step.find(" ")]
                step = int(step)
            else:
                raise ValueError("No step in fixed step?!")
            if step != span:
                raise ValueError(
                    "Parser currently only supports step == span, and they were not equal"
                )
            lastScore = None
            stop = start
            continue

        elif "Bed format" in row:
            mode = "bed"
            return _wiggle_with_bed_to_intervals(fo, "\t")
        if (
            not mode
        ):  # no mode set so far, perhaps set one now. Not 'continue', we need this row!
            if len(row.split("\t")) == 4:
                mode = "defined"
                # TODO: replace with dataframe csv reading in this case, should
                # be a bit faster
                # because the iterator reads in blocks...
                fo.seek(bytes_so_far - len(row), os.SEEK_SET)
                df = _wiggle_with_bed_to_intervals(fo, "\t")
                df.convert_type("chr", chromosome_mangler)
                return df
            elif len(row.split()) == 4:
                fo.seek(bytes_so_far - len(row), os.SEEK_SET)
                df = _wiggle_with_bed_to_intervals(fo, None)
                df.convert_type("chr", chromosome_mangler)
                return df
            else:
                raise ValueError(
                    "no step mode definied in file. add variablestep/fixedstep to %s"
                    % fo.name
                )
        if mode == "variable":
            row = row.split("\t")
            pos = int(row[0])
            score = float(row[1])
            if score != lastScore:
                if start is not None:
                    data["chr"].append(chromosome_mangler(chr))
                    data["start"].append(start)
                    data["stop"].append(stop)
                    data["score"].append(lastScore)
                lastScore = score
                start = pos
                stop = pos + span
            else:
                stop = pos + span
        elif mode == "fixed":
            score = float(row.strip())
            if score != lastScore:
                if lastScore:
                    data["chr"].append(chromosome_mangler(chr))
                    data["start"].append(start)
                    data["stop"].append(stop)
                    data["score"].append(lastScore)
                start = stop + 1
                lastScore = score
            stop += step
        elif mode == "defined":
            row = row.split("\t")
            chr = row[0]
            start = int(row[1])
            stop = int(row[2])
            score = float(row[3])
            data["chr"].append(chr)
            data["start"].append(start)
            data["stop"].append(stop)
            data["score"].append(score)
    if mode == "variable" or mode == "fixed":
        data["chr"].append(chromosome_mangler(chr))
        data["start"].append(start)
        data["stop"].append(stop)
        data["score"].append(lastScore)
    return pd.DataFrame(data)


def _wiggle_with_bed_to_intervals(file_handle, separator):
    print("calling _wiggle_with_bed_to_intervals")
    res = pd.read_csv.read(file_handle, header=None, sep=separator)
    res.rename_column("column0", "chr")
    res.rename_column("column1", "start")
    res.rename_column("column2", "stop")
    res.rename_column("column3", "score")
    return res

    pass



import hashlib
import collections
import itertools
import time
import pathlib
import subprocess

cmd = ["python", "setup.py", "docs"]
hashes = collections.defaultdict(lambda: "")


def get_hash(fn, second=False):
    try:
        with open(fn, "rb") as d:
            return hashlib.md5(d.read()).hexdigest()
    except FileNotFoundError:
        if not second:
            time.sleep(1)
            return get_hash(fn, True)
        else:
            raise


while True:
    rebuild = False
    files = ["docs/conf.py", "docs/_static/my-styles.css"]
    for fn in itertools.chain(
        files,
        pathlib.Path(".").glob("**/*.md"),
        pathlib.Path(".").glob("**/*.rst"),
        pathlib.Path("./src").glob("**/*.py"),
    ):
        fn = str(fn)
        new_hash = get_hash(fn)
        if hashes[fn] != new_hash:
            rebuild = True
        hashes[fn] = new_hash
    if rebuild:
        subprocess.check_call(cmd)



"""sequence/chipseq formats"""
import io
from .util import open_file, chunkify
import numpy
import pandas as pd
import os
import tempfile
import subprocess


def normalize_strand(x):
    if x == 1 or x == 0 or x == -1:
        return x
    if x == "+":
        return 1
    elif x == "-":
        return -1
    else:
        return 0


class BedEntry(object):
    __slots__ = ["refseq", "position", "length", "strand", "score", "name"]

    def __init__(self, chr, chrStart, chrEnd, name=None, strand=None, score=None):
        self.refseq = chr
        pos = int(chrStart)
        self.position = pos
        self.length = int(chrEnd) - pos
        if strand:
            self.strand = normalize_strand(strand)
        else:
            self.strand = normalize_strand(self.length > 0)
        self.score = numpy.nan
        if name is None:
            name = "Noname"
        self.name = name
        if score is not None:
            self.score = score

    def get_read_length(self):
        return self.length

    def __len__(self):
        return self.length

    def __repr__(self):
        return "BedEntry(%s, %i, %i)" % (
            repr(self.refseq),
            self.position,
            self.position + self.length,
        )

    def __str__(self):
        print(
            "Bed Entry chr=%s, start=%i, length=%i, strand=%s, score=%s, name=%s"
            % (
                self.refseq,
                self.position,
                self.length,
                self.strand,
                self.score,
                self.name,
            )
        )


def read_bed(filenameOrFileObject, report_progress=False):
    res = []
    for e in read_bed_iterator(filenameOrFileObject, report_progress):
        res.append(e)
    return res


def read_bed_iterator(filenameOrFileObject, report_progress=False):
    fo = open_file(filenameOrFileObject, "rb")
    for row in chunkify(fo, b"\n"):
        if row.startswith(b"track"):
            trackInfo = row
        elif row.startswith(b"#"):  # not really a comment character...
            continue
        else:
            try:
                if not row:
                    continue
                row = row.split(b"\t")
                e = BedEntry(row[0], row[1], row[2])  # bed does start at 0
                try:
                    e.name = row[3]
                    e.score = float(row[4])
                except IndexError:
                    pass
                except ValueError:
                    pass
                try:
                    e.strand = normalize_strand(row[5])
                except IndexError:
                    pass
                yield e
            except Exception as e:
                raise ValueError("Could not parse row: %s" % row)


def write_bed_header(file_handle, track_name):
    file_handle.write(
        ('track name="%s" description="" useScore=0\n' % track_name).encode("utf-8")
    )


def write_bed_entry_short(file_handle, entry, name_to_chromosome_lookup_or_none=None):
    if name_to_chromosome_lookup_or_none:
        chr = name_to_chromosome_lookup_or_none[entry.refseq]
    else:
        chr = entry.refseq
    out = [
        chr,  # chromosome
        entry.position,  # start
        entry.position + len(entry),  # end
    ]
    file_handle.write("\t".join((str(x) for x in out)) + "\n")


def write_bed_entry_long(file_handle, entry, name_to_chromosome_lookup_or_none=None):
    if name_to_chromosome_lookup_or_none:
        chr = name_to_chromosome_lookup_or_none[entry.refseq]
    else:
        chr = entry.refseq
    out = [
        chr,  # chromosome
        entry.position,  # start
        entry.position + len(entry),  # end
        entry.name,
        "." if numpy.isnan(entry.score) else entry.score,
        "+" if entry.strand == 1 else "-" if entry.strand == -1 else ".",
        entry.position,  # this start
        entry.position + len(entry),
    ]
    file_handle.write(b"\t".join((str(x).encode("utf-8") for x in out)) + b"\n")


def write_bed(
    filenameOrFileObject,
    reads,
    name_to_chromosome_lookup_or_none,
    track_name,
    include_header=True,
    minimal=False,
):
    fo = open_file(filenameOrFileObject, "wb")
    if include_header:
        write_bed_header(fo, track_name)
    if minimal:
        write_method = write_bed_entry_short
    else:
        write_method = write_bed_entry_long
    for read in reads:
        write_method(fo, read, name_to_chromosome_lookup_or_none)
    if fo != filenameOrFileObject:
        fo.close()


def bed_to_bigbed(
    input_bed_filename, output_filename, chromosome_lengths, already_sorted=False
):
    """Convert an existing bed file into bigbed. Chromosome lengths is a dictionary"""
    from mbf_externals.kent import BedToBigBed

    algo = BedToBigBed()
    algo.store.unpack_version(algo.name, algo.version)

    chrom_sizes_file = tempfile.NamedTemporaryFile(suffix=".sizes")
    for chr, length in sorted(chromosome_lengths.items()):
        chrom_sizes_file.write(("%s\t%i\n" % (chr, length)).encode("utf-8"))
    chrom_sizes_file.flush()

    input = open(input_bed_filename, "rb")
    if not already_sorted:
        missing_chroms = set()
        out = list()
        chrom_lenghts = chromosome_lengths
        for line in input:
            line = line.strip()
            if line and not line.startswith(b"#") and not line.startswith(b"track"):
                line = line.split(b"\t")
                try:
                    chr = line[0].decode("utf-8")
                    start = min(max(0, int(line[1])), chrom_lenghts[chr] - 1)
                    stop = min(int(line[2]), chrom_lenghts[chr] - 1)
                    strand = None
                    name = None
                    try:
                        name = line[3].decode("utf-8")
                        strand = line[5].decode("utf-8")
                    except IndexError:
                        if name is None:
                            name = "."
                        if strand is None:
                            strand = "."
                    out.append(
                        (chr, min(start, stop), max(start, stop), name, 0, strand)
                    )
                except KeyError:
                    missing_chroms.add(chr)
        if missing_chroms:
            print("could not find the following chromosomes %s" % (missing_chroms,))
        out.sort()
        inputtf = tempfile.NamedTemporaryFile()
        for chr, start, stop, name, score, strand in out:
            inputtf.write(
                (
                    "%s\t%s\t%s\t%s\t%s\t%s\n" % (chr, start, stop, name, score, strand)
                ).encode("utf-8")
            )
        inputtf.flush()
        inputtf.seek(0, 0)
        input.close()
    else:
        inputtf = input
    in_name = os.path.abspath(inputtf.name)
    cmd = [
        str(algo.path / "bedToBigBed"),
        "-tab",
        in_name,
        os.path.abspath(chrom_sizes_file.name),
        os.path.abspath(output_filename),
        "-type=bed6",
    ]
    print(cmd)
    p = subprocess.Popen(cmd)
    p.communicate()
    inputtf.close()
    chrom_sizes_file.close()
    if p.returncode != 0:
        raise ValueError("bedToBigBed returned an error code")


def write_bigbed(
    input_dataframe_or_list_of_bed_entries, output_filename, chromosome_lengths
):
    """Take either a list of BedEntry objects, a dataframe with the slot names of BedEntry objects, or
    a DataFrame with {chr, start, stop, strand, score, name}.
    Either way, strand needs to be one of -1,0,1

    """
    if not isinstance(input_dataframe_or_list_of_bed_entries, pd.DataFrame):
        r = {}
        for column in ["refseq", "position", "length", "strand", "score", "name"]:
            r[column] = []
            for e in input_dataframe_or_list_of_bed_entries:
                r[column].append(getattr(e, column))
        df = pd.DataFrame(r)
    else:
        df = input_dataframe_or_list_of_bed_entries

    if "chr" in df.columns and "start" in df.columns and "stop" in df.columns:
        df = df.sort_values(["chr", "start"], ascending=[True, True])
    elif "refseq" in df.columns and "position" in df.columns and "length" in df.columns:
        df = df.rename(columns={"refseq": "chr", "position": "start"})
        df = df.assign(stop=df["start"] + df["length"])
        df = df.sort_values(["chr", "start"], ascending=[True, True])
    else:
        raise ValueError(
            "This dataframe did not contain the necessary bed columns (either (chr, start, stop) or (refseq, position, length)"
        )
    if "strand" not in df.columns:
        df = df.assign("strand", 0)
    if "name" not in df.columns:
        df = df.assign("name", "Noname")
    if "score" not in df.columns:
        df = df.assign("score", numpy.nan)
    of = tempfile.NamedTemporaryFile(suffix=".bed")
    for dummy_idx, row in df.iterrows():
        output = "%s\t%i\t%i\t%s\t%s\t%s\n" % (
            row["chr"],
            row["start"],
            row["stop"],
            row["name"],
            ("%f" % (row["score"],)) if (row["score"] is not None) else ".",
            ("+" if row["strand"] == 1 else "-" if row["strand"] == -1 else "+"),
        )
        output = output.encode("utf-8")
        of.write(output)
    of.flush()
    bed_to_bigbed(of.name, output_filename, chromosome_lengths)
    of.close()


def read_bigbed(filename, chromosome_lengths, chromosome_mangler=lambda x: x):
    import pyBigWig

    bb = pyBigWig.open(filename)
    chr_lengths = chromosome_lengths
    data = {"chr": [], "start": [], "stop": [], "strand": [], "name": []}
    for chr in chr_lengths:
        it = bb.entries(chromosome_mangler(chr), 0, chr_lengths[chr])
        if it is None:  # no such chromosome. Tolerable if it's a contig or such.
            # If none of the chromosome names match,
            # we raise later because of an empty big file.
            continue
        for entry in it:
            data["chr"].append(chr)
            data["start"].append(entry[0])
            data["stop"].append(entry[1])
            more = entry[2].split("\t")
            strand = more[2]
            data["strand"].append(1 if strand == "+" else -1 if strand == "-" else 0)
            data["name"].append(more[0])
    bb.close()
    return pd.DataFrame(data)



from mbf_fileformats.util import chunkify, open_file


def test_open_file():
    import gzip
    import tempfile
    import bz2

    tf = tempfile.TemporaryFile()
    assert open_file(tf) is tf

    tf2 = tempfile.NamedTemporaryFile(suffix=".gz", mode="w")
    g = gzip.GzipFile(tf2.name, "w")
    g.write(b"hello")
    g.close()
    tf2.flush()
    assert open_file(tf2.name).read() == b"hello"

    tf3 = tempfile.NamedTemporaryFile(suffix=".bz2", mode="w")
    b = bz2.BZ2File(tf3.name, "w")
    b.write(b"world")
    b.close()
    tf3.flush()
    assert open_file(tf3.name).read() == b"world"


def test_chunkify():
    import tempfile

    tf = tempfile.TemporaryFile("w+")
    tf.write("hello world")
    tf.flush()
    tf.seek(0, 0)
    c = list(chunkify(tf, " ", 2))
    assert c == ["hello", "world"]



from mbf_externals.util import to_bytes
from .util import open_file
from .bed import normalize_strand, read_bed


def _mapGFF(row):
    res = {
        'seqname': row[0],
        'source': row[1],
        'feature': row[2],
        'start': int(row[3]) if row[3] != '.' else 0,
        'end': int(row[4]),
        'score': row[5],
        'strand': normalize_strand(row[6]),
        'frame': row[7],
        'attributes': {},
        'comment': ''
    }
    attributes = row[8]
    comment = ''
    if attributes.find(b'#') != -1:
        comment = attributes[attributes.find(b'#') + 1:]
        attributes = attributes[:attributes.find(b'#')]
    res['comment'] = comment
    for x in attributes.split(b';'):
        if x.strip():
            if b'=' in x:
                x = x.split(b"=")
            else:
                x = x.split()
            res['attributes'][x[0].decode('utf-8')] = [y.decode('utf-8') for y in x[1:]]
    return res


def gffToDict(filename, comment_char=None):
    comment_char = to_bytes(comment_char)
    o = open_file(filename)
    rows = o.readlines()
    rows = (x.strip() for x in rows)
    rows = (x.split(b"\t") for x in rows if x and (comment_char is None or x[0] != comment_char[0]))
    res = [_mapGFF(x) for x in rows]
    return res


def dictsToGFF(gffDicts, filename):
    rows = []
    for entry in gffDicts:
        if 'score' in entry:
            score = entry['score']
        else:
            score = '.'
        if 'attributes' in entry:
            attributes = entry['attributes']
            if isinstance(attributes, dict):
                attributes_str = []
                for key, value in attributes.items():
                    attributes_str.append("%s=%s" % (key, value))
                attributes_str = ";".join(attributes_str)
        else:
            attributes_str = ''
        if entry['strand'] not in ('+', '-', '.'):
            raise ValueError("invalid strand: %s" % entry['strand'])
        try:
            entry['frame'] = int(entry['frame'])
            if entry['frame'] not in (0, 1, 2):
                raise ValueError()
        except ValueError:
            raise ValueError("invalid frame: %s " % entry['frame'])
        row = "\t".join((entry['seqname'], entry['source'], entry['feature'], str(int(entry['start'])), str(int(entry['end'])),
                         str(score), entry['strand'], str(entry['frame']), attributes_str))
        rows.append(row)
    o = open(filename, 'wb')
    o.write("\n".join(rows))
    o.close()


class GFF3:

    def escape(self, str):
        if str is None:
            return '.'
        escape = '"\t\n\r=;'
        for k in escape:
            str = str.replace(k, '%' + '%2x' % ord(k))
        return str

    def format_attributes(self, attributes):
        if attributes is None:
            return '.'
        if isinstance(attributes, dict):
            attributes = list(attributes.items())
        valid_attributes = [ 'Name', 'Alias', 'Parent', 'Target', 'Gap', 'Derives_from', 'Note', 'Dbxref']
        res = []
        for id, value in attributes:
        #    if not id in valid_attributes and id != id.lower(): #lower case names are not reserved
         #       raise ValueError("Not a valid tag: %s" % id)
            res.append('%s=%s' % (self.escape(id), self.escape(value)))
        return ";".join(res)


    def dump_row(self, file_handle, seqid = None, source = None, type = None, start = None, end = None, score = None, strand = None, phase = None, attributes = None):
        file_handle.write("\t".join((
            self.escape(seqid),
            self.escape(source),
            self.escape(type),
            self.escape(start),
            self.escape(end),
            self.escape(score),
            self.escape(strand),
            self.escape(phase),
            self.format_attributes(attributes))))
        file_handle.write("\n")


def bed_to_gff(input_filename_or_handle, output_filename_or_handle, feature_type, source = None, show_progress= False):
    entries = read_bed(input_filename_or_handle, report_progress = show_progress)
    output_file_handle = open_file(output_filename_or_handle, 'wb')
    gff = GFF3()
    for e in entries:
        gff.dump_row(output_file_handle, e.refseq, source=source, type=feature_type, start = str(e.position), end = str(e.position + e.length), strand='.')
    output_file_handle.close()



from pathlib import Path


def open_file(fileNameOrHandle, mode="rb"):
    """Transparently open compressed or uncompressed files"""
    if hasattr(fileNameOrHandle, "read"):
        return fileNameOrHandle
    elif isinstance(fileNameOrHandle, Path):
        fileNameOrHandle = str(fileNameOrHandle)
    if fileNameOrHandle.endswith(".gz"):
        import gzip

        return gzip.GzipFile(fileNameOrHandle, mode)
    elif fileNameOrHandle.endswith(".bz2"):
        import bz2

        return bz2.BZ2File(fileNameOrHandle, mode)
    else:
        return open(fileNameOrHandle, mode)


def chunkify(handle, separator, block_size=None):
    """take a file handle and split it at separator, reading in efficently in 50 mb blocks or so"""
    if block_size is None:
        block_size = 50 * 1024 * 1024
    chunk = handle.read(block_size)
    chunk = chunk.split(separator)
    while True:
        for k in chunk[:-1]:
            yield k
        next = handle.read(block_size)
        if next:
            chunk = chunk[-1] + next
            chunk = chunk.split(separator)
        else:
            yield chunk[-1]
            break


def pathify(output_filename, default, create_parents=True):
    if output_filename is None:
        res = Path(default)
    else:
        res = Path(output_filename)

    if create_parents:
        res.parent.mkdir(exist_ok=True)
    return res




import pandas as pd
import numpy as np


def read_prism_tables(filename):
    try:
        import BeautifulSoup as bs
        import HTMLParser

        html_decode = HTMLParser.unescape
    except ImportError:
        import bs4 as bs

        # import html
        html_decode = lambda x: x  # bs4 already does this
    with open(filename) as op:
        x = bs.BeautifulSoup(op.read(), "lxml")
    result = []
    for t in x.findAll("table"):
        titles = [html_decode(title.text) for title in t.findAll("title")]
        columns = []
        max_length = 0
        for subcolumn in t.findAll("subcolumn"):
            c = []
            float_count = 0
            for d in subcolumn.findAll("d"):
                dt = html_decode(d.text)
                if dt == "":
                    dt = np.nan
                try:
                    dt = float(dt)
                    float_count += 1
                except ValueError:
                    if dt.count(",") == 1 and dt.count(".") == 0:
                        try:
                            dt = float(dt.replace(",", "."))
                            float_count += 1
                        except ValueError:
                            pass
                c.append(dt)
            if float_count <= 5:
                c = ["" if isinstance(x, float) and np.isnan(x) else x for x in c]
            columns.append(c)
            max_length = max(max_length, len(c))
        for c in columns:
            while len(c) < max_length:
                c.append(np.nan)
        df = pd.DataFrame(dict(zip(titles, columns)))[titles]
        result.append(df)
    return result
